"""Definitions subpackage containing shared data structures used throughout reviser."""

from .abstracts import DataWrapper
from .abstracts import Specification
from .aws import AwsConnection
from .configurations import Bundle
from .configurations import Configuration
from .configurations import Dependency
from .configurations import EnvironmentVariable
from .configurations import PipDependency
from .configurations import PipperDependency
from .configurations import PoetryCommandDependency
from .configurations import PoetryDependency
from .configurations import Target
from .configurations import UvCommandDependency
from .configurations import UvDependency
from .contexts import Context
from .contexts import SelectedTargets
from .enumerations import RUNTIME_VERSION
from .enumerations import DefaultFile
from .enumerations import DependencyType
from .enumerations import TargetType
from .infomatics import LambdaFunction
from .infomatics import LambdaLayer
from .infomatics import LambdaLayerReference
from .infomatics import PublishedLayer
from .selections import Selection

__all__ = [
    "AwsConnection",
    "Bundle",
    "Configuration",
    "Context",
    "DataWrapper",
    "DefaultFile",
    "Dependency",
    "DependencyType",
    "EnvironmentVariable",
    "LambdaFunction",
    "LambdaLayer",
    "LambdaLayerReference",
    "PipDependency",
    "PipperDependency",
    "PoetryCommandDependency",
    "PoetryDependency",
    "PublishedLayer",
    "RUNTIME_VERSION",
    "SelectedTargets",
    "Selection",
    "Specification",
    "Target",
    "TargetType",
    "UvCommandDependency",
    "UvDependency",
]



"""Dependency configuration data structures and IO module."""

import dataclasses
import json
import pathlib
import subprocess
import sys
import tempfile
import typing

from reviser import utils
from reviser.definitions import abstracts
from reviser.definitions import aws
from reviser.definitions import configurations
from reviser.definitions import enumerations


def parse_dependencies(
    directory: pathlib.Path,
    connection: "aws.AwsConnection",
    dependency_data: typing.List[typing.Dict[str, typing.Any]],
    is_shared: bool = False,
):
    """Parse configuration dependency data into a tuple of Dependency objects."""
    output = []

    dependency_types = enumerations.DependencyType
    mappings = typing.cast(
        typing.Dict[str, typing.Any],
        {
            dependency_types.PIP.value: PipDependency,
            dependency_types.PIPPER.value: PipperDependency,
            dependency_types.POETRY.value: PoetryDependency,
            dependency_types.POETRY_COMMAND.value: PoetryCommandDependency,
            dependency_types.UV.value: UvDependency,
            dependency_types.UV_COMMAND.value: UvCommandDependency,
        },
    )
    for data in dependency_data:
        kind = data.get("kind", dependency_types.PIP.value)
        constructor = typing.cast(
            typing.Any,
            mappings.get(kind, configurations.PipDependency),
        )
        output.append(
            constructor(
                directory=directory,
                data=data,
                connection=connection,
                is_shared=is_shared,
            )
        )

    return tuple(output)


@dataclasses.dataclass(frozen=True)
class DependencyGroup(abstracts.Specification):
    """Group of package dependencies."""

    configuration: "configurations.Configuration"
    target: typing.Optional["configurations.Target"] = None
    name: typing.Optional[str] = None

    @property
    def is_shared(self) -> bool:
        """Whether the dependency group is shared, or specific to a target."""
        return self.target is None

    @property
    def site_packages_directory(self) -> pathlib.Path:
        """Get the temporary location where site packages will be installed."""
        if self.target:
            return self.target.site_packages_directory
        return pathlib.Path(tempfile.gettempdir()).joinpath(self.uuid)

    @property
    def sources(self) -> typing.Tuple["Dependency", ...]:
        """Parse configuration dependency data into a tuple of Dependency objects."""
        output = []

        dependency_types = enumerations.DependencyType
        mappings = typing.cast(
            typing.Dict[str, typing.Any],
            {
                dependency_types.PIP.value: PipDependency,
                dependency_types.PIPPER.value: PipperDependency,
                dependency_types.POETRY.value: PoetryDependency,
                dependency_types.POETRY_COMMAND.value: PoetryCommandDependency,
                dependency_types.UV.value: UvDependency,
                dependency_types.UV_COMMAND.value: UvCommandDependency,
            },
        )
        for data in self.get_as_list("sources") or []:
            kind = data.get("kind", dependency_types.PIP.value)
            constructor = typing.cast(
                typing.Any, mappings.get(kind, configurations.PipDependency)
            )
            output.append(
                constructor(
                    directory=self.directory,
                    data=data,
                    connection=self.connection,
                    group=self,
                )
            )

        return tuple(output)

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {"sources": [d.serialize() for d in self.sources]}


@dataclasses.dataclass(frozen=True)
class Dependency(abstracts.Specification):
    """Package dependency data structure."""

    group: "DependencyGroup"

    @property
    def kind(self) -> "enumerations.DependencyType":
        """Get the kind of dependency."""
        return enumerations.DependencyType(
            value=self.get(
                "kind",
                default=enumerations.DependencyType.PIP.value,
            )
        )

    @property
    def file(self) -> typing.Optional[pathlib.Path]:
        """Get the optionally-specified path used to load dependencies."""
        if value := self.get("file"):
            return self.directory.joinpath(value).absolute()

        if self.get("packages") is not None:
            return None

        default = self.directory.joinpath(
            enumerations.DefaultFile.from_dependency_type(self.kind).value
        )
        return default if default.exists() else None

    @property
    def packages(self) -> typing.List[str]:
        """Get the explicitly defined packages."""
        return self.get_first_as_list(["packages"], ["package"], default=[])

    @property
    def skip_packages(self) -> typing.List[str]:
        """Get packages that should be skipped during the installation process."""
        return self.get_first_as_list(["skips"], ["skip"], default=[])

    @property
    def arguments(self) -> typing.Dict[str, str]:
        """Gets arguments that should be passed during the installation process."""
        return self.get_first_as_dict(["arguments"], default={})

    @property
    def command_args(self) -> typing.List[str]:
        """List command arguments for a command-based execution."""
        return self.get_as_list("args") or []

    def get_package_names(self) -> typing.List[str]:
        """List package names to install from the various package sources."""
        return []

    def execute_command(self):
        """Execute installation command specified in the configuration."""
        pass


@dataclasses.dataclass(frozen=True)
class PipDependency(Dependency):
    """Pip package dependency data structure."""

    def get_package_names(self) -> typing.List[str]:
        """List of package names to install collected from the various sources."""
        packages = self.packages.copy()
        if self.file:
            packages += [
                item
                for line in (self.file.read_text() or "").split("\n")
                if (item := line.strip())
            ]
        return [
            p
            for p in packages
            if utils.extract_package_name(p) not in self.skip_packages
        ]

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
            "packages": self.get_package_names(),
        }


@dataclasses.dataclass(frozen=True)
class PipperDependency(Dependency):
    """Pipper package dependency data structure."""

    @property
    def prefix(self) -> typing.Optional[str]:
        """Get the custom S3 key prefix in the bucket where the repository resides."""
        return self.get("prefix")

    @property
    def bucket(self) -> typing.Optional[str]:
        """Get the S3 bucket where the pipper repository resides."""
        return utils.get_matching_bucket(
            buckets=self.get_first(["buckets"], ["bucket"]),
            aws_region=self.group.configuration.aws_region,
            aws_account_id=self.connection.aws_account_id,
        )

    def get_package_data(self) -> typing.Optional[dict]:
        """Get the  data stored in the specified package file."""
        if self.file:
            return json.loads(self.file.read_text())
        return None

    def get_package_names(self) -> typing.List[str]:
        """List collected package names to install from various definition sources."""
        packages = self.packages.copy()
        if data := self.get_package_data():
            packages += data.get("dependencies") or []
        return [
            p
            for p in packages
            if utils.extract_package_name(p) not in self.skip_packages
        ]

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
            "prefix": self.prefix,
            "bucket": self.bucket,
            "packages": self.get_package_names(),
        }


def _find_poetry_executable() -> str:
    """Find the poetry executable path, or raise an error if not found."""
    directories = [
        pathlib.Path("~/.local/bin").expanduser(),
        pathlib.Path(sys.prefix).joinpath("bin"),
        pathlib.Path(sys.executable).parent,
    ]
    directories = [d for d in directories if d.exists()]

    finder = (
        str(p)
        for d in directories
        for p in d.iterdir()
        if p.name == "poetry" or p.name.startswith("poetry.")
    )
    try:
        return next(finder)
    except StopIteration:
        raise RuntimeError(
            "Unable to find poetry installation in current Python environment."
        )


@dataclasses.dataclass(frozen=True)
class PoetryDependency(Dependency):
    """Poetry package dependency data structure."""

    @property
    def extras(self) -> typing.List[str]:
        """List extra packages to install."""
        return self.get_as_list("extras", default=self.get_as_list("extra")) or []

    def get_package_names(self) -> typing.List[str]:
        """List collected package names to install from various definition sources."""
        packages = self.packages.copy()
        if not self.file:
            return packages

        executable = _find_poetry_executable()

        # Ensure that a lock file exists and is up-to-date before proceeding.
        command = [executable, "lock"]
        subprocess.run(command, stdout=subprocess.PIPE, check=True)

        command = [
            executable,
            "export",
            "--format=requirements.txt",
            "--without-hashes",
        ]

        for group in self.extras:
            command.append(f"--extras={group}")

        result = subprocess.run(command, stdout=subprocess.PIPE, check=True)

        packages += [
            item
            for line in result.stdout.decode().strip().split("\n")
            if (item := line.split(";")[0].strip())
        ]

        return [
            p
            for p in packages
            if utils.extract_package_name(p) not in self.skip_packages
        ]

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
            "packages": self.get_package_names(),
        }


def _find_uv_executable() -> str:
    """Find the uv executable path, or raise an error if not found."""
    result = subprocess.run(
        ["uv", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    if result.returncode == 0:
        return "uv"

    directories = [
        pathlib.Path("~/.local/bin").expanduser(),
        pathlib.Path(sys.prefix).joinpath("bin"),
        pathlib.Path(sys.executable).parent,
    ]
    directories = [d for d in directories if d.exists()]

    finder = (
        str(p)
        for d in directories
        for p in d.iterdir()
        if p.name == "uv" or p.name.startswith("uv.")
    )
    try:
        return next(finder)
    except StopIteration:
        raise RuntimeError(
            "Unable to find uv installation in current Python environment."
        )


@dataclasses.dataclass(frozen=True)
class UvDependency(Dependency):
    """UV package dependency data structure."""

    @property
    def extras(self) -> typing.List[str]:
        """List extra packages to install."""
        return self.get_as_list("extras", default=self.get_as_list("extra")) or []

    def get_package_names(self) -> typing.List[str]:
        """List collected package names to install from various definition sources."""
        packages = self.packages.copy()
        if not self.file:
            return packages

        executable = _find_uv_executable()

        # Ensure that a lock file exists and is up-to-date before proceeding.
        command = [executable, "lock"]
        subprocess.run(command, stdout=subprocess.PIPE, check=True)

        command = [
            executable,
            "pip",
            "--format=requirements.txt",
            "--no-hashes",
            "--no-annotate",
        ]

        for group in self.extras:
            command.append(f"--extras={group}")

        result = subprocess.run(command, stdout=subprocess.PIPE, check=True)

        packages += [
            item
            for line in result.stdout.decode().strip().split("\n")
            if (item := line.split(";")[0].strip())
        ]

        return [
            p
            for p in packages
            if utils.extract_package_name(p) not in self.skip_packages
        ]

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
            "packages": self.get_package_names(),
        }


@dataclasses.dataclass(frozen=True)
class UvCommandDependency(Dependency):
    """UV command-based package dependency data structure."""

    @property
    def extras(self) -> typing.List[str]:
        """List extra packages to install."""
        return []

    def get_package_names(self) -> typing.List[str]:
        """List collected package names to install from various definition sources."""
        return []

    def execute_command(self):
        """Execute installation command specified in the configuration."""
        executable = _find_uv_executable()
        command = [executable, *self.command_args]
        subprocess.run(command, stdout=subprocess.PIPE, check=True)

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
        }


@dataclasses.dataclass(frozen=True)
class PoetryCommandDependency(Dependency):
    """Poetry command-based package dependency data structure."""

    @property
    def extras(self) -> typing.List[str]:
        """List extra packages to install."""
        return []

    def get_package_names(self) -> typing.List[str]:
        """List collected package names to install from various definition sources."""
        return []

    def execute_command(self):
        """Execute installation command specified in the configuration."""
        executable = _find_poetry_executable()
        command = [executable, *self.command_args]
        subprocess.run(command, stdout=subprocess.PIPE, check=True)

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "kind": self.kind.value,
        }



"""Data structure and IO module for function and layer targets."""

import dataclasses
import fnmatch
import pathlib
import re
import tempfile
import typing

from botocore.client import BaseClient

from reviser import utils
from reviser.definitions import abstracts
from reviser.definitions import configurations
from reviser.definitions import enumerations
from reviser.definitions import selections


def _is_selected_match(
    target: "Target",
    name: str,
):
    """Determine whether the name matches the current selection set on the target."""
    if not target.selection:
        return True

    if exacts := getattr(target.selection, f"{target.kind.value}_names"):
        return name in exacts

    needles = getattr(target.selection, f"{target.kind.value}_needles")
    any_needles = bool(target.selection.function_needles) or bool(
        target.selection.layer_needles
    )

    return (
        # Match everything if `--all` was specified.
        target.selection.bundle_all
        # No needles specified.
        or not any_needles
        # Identical match.
        or name in needles
        # Shell-style wildcard matches.
        or any([fnmatch.fnmatch(name, n) for n in needles])
        # Partial character matches.
        or any([n.lower() in name.lower() for n in needles])
    )


@dataclasses.dataclass(frozen=True)
class Target(abstracts.Specification):
    """
    Data structure for defining a target.

    Function or layer definition data structure where a target can
    represent one or more functions or one or more layers in the same
    definition for cases where function/layer duplication is desirable.
    A target, however, is of a homogeneous kind, meaning it will only
    be a function or layer.
    """

    configuration: "configurations.Configuration"
    selection: typing.Optional["selections.Selection"] = None

    @property
    def aws_region(self) -> str:
        """Get the AWS region name where this target resides."""
        return self.get("region", default=self.configuration.aws_region)

    @property
    def bucket(self) -> typing.Optional[str]:
        """Get the  bucket to use for uploading to S3."""
        return utils.get_matching_bucket(
            buckets=self.get_first(["buckets"], ["bucket"]),
            aws_region=self.aws_region,
            aws_account_id=self.connection.aws_account_id,
            default_bucket=self.configuration.bucket,
        )

    @property
    def bundle(self) -> "configurations.Bundle":
        """Get the bundle object associated with this target configuration."""
        return configurations.Bundle(
            directory=self.directory,
            data=self.get("bundle", default={}),
            connection=self.connection,
            target=self,
        )

    @property
    def image(self) -> "configurations.Image":
        """Get the image object associated with this target configuration."""
        return configurations.Image(
            directory=self.directory,
            data=self.get("image", default={}),
            connection=self.connection,
            target=self,
        )

    @property
    def layer_attachments(self) -> typing.List["configurations.AttachedLayer"]:
        """
        List layers that should be attached to functions in this target.

        This will always be an empty list for layer targets.
        """
        if self.kind == enumerations.TargetType.LAYER:
            return []
        values = self.get_first_as_list(["layers"], ["layer"], default=[])
        values = [{"name": v} if isinstance(v, str) else v for v in values]
        return [
            configurations.AttachedLayer(
                directory=self.directory,
                data=v,
                connection=self.connection,
                target=self,
            )
            for v in values
        ]

    @property
    def variables(self) -> typing.List["configurations.EnvironmentVariable"]:
        """
        List environment variables that will be applied during function updates.

        This will always be an empty list for layer targets that don't support
        environment variables.
        """
        if self.kind == enumerations.TargetType.LAYER:
            return []
        values = self.get_first_as_list(["variables"], ["variable"], default=[])
        values = [{"arg": v} if isinstance(v, str) else v for v in values]
        return [
            configurations.EnvironmentVariable(
                directory=self.directory,
                data=v,
                connection=self.connection,
                target=self,
            )
            for v in values
        ]

    @property
    def ignores(self) -> typing.List[str]:
        """
        List items to skip when modifying function configurations.

        These are the items that should be skipped when modifying function
        configurations during the deployment process.
        """
        ignores = self.get_first_as_list(["ignores"], ["ignore"], default=[])
        return [i.lower() for i in ignores]

    @property
    def bundle_directory(self) -> pathlib.Path:
        """Get the temporary location where the target will be assembled."""
        return pathlib.Path(tempfile.gettempdir()).joinpath(self.uuid)

    @property
    def bundle_zip_path(self) -> pathlib.Path:
        """Get the temporary location where the target zip bundle will be saved."""
        return pathlib.Path(tempfile.gettempdir()).joinpath(f"{self.uuid}.zip")

    @property
    def site_packages_directory(self) -> pathlib.Path:
        """Get the temporary location where site packages will be installed."""
        if self.kind == enumerations.TargetType.LAYER:
            return self.bundle_directory.joinpath("python")
        return self.bundle_directory.joinpath("site_packages")

    @property
    def kind(self) -> "enumerations.TargetType":
        """Get the target type as one of 'function' or 'layer'."""
        # noinspection PyArgumentList
        return enumerations.TargetType(
            value=self.get(
                "kind",
                default=enumerations.TargetType.FUNCTION.value,
            )
        )

    @property
    def names(self) -> typing.List[str]:
        """Get the names of the targets associated with this definition."""
        names = self.get_first_as_list(["names"], ["name"], default=[])
        return [str(n) for n in names if n and _is_selected_match(self, n)]

    @property
    def timeout(self) -> typing.Optional[int]:
        """
        Get the seconds after which the function stops executing.

        This will always be None for layers.
        """
        value = self.get("timeout")
        if not value or self.kind == enumerations.TargetType.LAYER:
            return None

        try:
            return int(value)
        except ValueError:
            regex = re.compile(r"^(?P<value>\d+).*")
            if (match := regex.match(value)) is not None:
                return int(match.groupdict()["value"])
            return None

    @property
    def memory(self) -> typing.Optional[int]:
        """
        Get the MB available to the function when executing.

        This will always be None for layers.
        """
        value = self.get("memory")
        if not value or self.kind == enumerations.TargetType.LAYER:
            return None

        try:
            return int(value)
        except ValueError:
            regex = re.compile(r"^(?P<value>\d+).*")
            if (match := regex.match(value)) is not None:
                return int(match.groupdict()["value"])
            return None

    @property
    def dependencies(self) -> "configurations.DependencyGroup":
        """Get the dependencies for this function definition."""
        dependency_data = self.get("dependencies")
        if isinstance(dependency_data, str):
            return self.configuration.shared_dependencies[dependency_data]

        if not isinstance(dependency_data, dict):
            # Mutate the dependency data into a dictionary if not already done.
            dependency_data = {"sources": dependency_data or []}
            self.data["dependencies"] = dependency_data

        return configurations.DependencyGroup(
            directory=self.directory,
            data=dependency_data,
            connection=self.connection,
            configuration=self.configuration,
            target=self,
        )

    def ignores_any(self, *args: str) -> bool:
        """Determine if any of the specified args appear in the ignores list."""
        ignores = self.ignores
        finder = (True for a in args if a in ignores)
        return next(finder, False)

    def client(self, service_name: str) -> BaseClient:
        """Create a boto3 client configured for use with this target."""
        return self.connection.client(service_name, region_name=self.aws_region)

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        if self.kind == enumerations.TargetType.FUNCTION:
            values = {
                "layers": [a.serialize() for a in self.layer_attachments],
                "variables": [v.serialize() for v in self.variables],
                "memory": self.memory,
                "timeout": self.timeout,
                "ignores": self.ignores,
            }
        else:
            values = {}

        return {
            "kind": self.kind.value,
            "names": self.names,
            "region": self.aws_region,
            "bundle_directory": str(self.bundle_directory),
            "bundle_zip_path": str(self.bundle_zip_path),
            "site_packages_directory": str(self.site_packages_directory),
            "bundle": self.bundle.serialize(),
            "image": self.image.serialize(),
            "dependencies": self.dependencies.serialize(),
            **values,
        }
